'''
Python script for the paper:

Lanfear R, Bromham L (2008) Statistical Tests Between Competing Hypotheses of Hox Cluster Evolution. Systematic Biology 57(5):1-11

You can get the PDF from here:

http://www.tempoandmode.com/people/robert-lanfear/

or email me at:

rob.lanfear@anu.edu.au

for the shiny version from systbiol

***********************************************************************
Description

This script is designed to take as input a bunch of nexus trees (i.e. from 
an output file of mrBayes3.1.2, and then process each tree, to work out whether
it supports a 2, 3, 4 or whatever protohox gene hypothesis.

You will need to have BioPython installed for it to work, you can find Biopython at:
http://biopython.org/wiki/Main_Page

NOTE: this script will ONLY work if ALL of your outgroup are ALWAYS recovered as 
monophyletic in your analysis (e.g. if every tree in your bayesian sample of trees
has the outgroup as monophyletic. If this isn't the case, you'll need to either
change the outgroup, re-write the script, or constrain the outgroup to be monophyletic.
However, if it isn't the case, you should probably be using a different outgroup
anyway.

The script is written to the best of my ability, i.e. very badly. I have no intention
of making it any better, as it does its job in a reasonable amount of time. 


Rob Lanfear, October 2008

************************************************************************
INPUT

A tree file from MrBayes3.1.2 generated by using the 'sumt' command. 

This file must have exactly the same sequence names as the example file provided, this is becuase 
I am a bad programmer, and everything, including sequence names, is hard-coded below.

OUTPUT

process_trees.out

lists all the hypotheses encountered in the tree file, derived by literally interpreting 
each of the trees in the output file. It tells you how big the minimal protohox cluster would have
been for a given hypothesis, and how many instances of Hox and ParaHox gene loss you'd have to infer
for that topology. It also gives the posterior probability of each hypothesis, calculated by summing
the posterior probability of each tree which matched that hypothesis in the trees file. You also get the 
posterior probability of rejecting the 'Tandem Duplication' model (see paper), listed after the word
'True' in the output file.

NB I used a slightly altered version of MrBayes 3.1.2 to generate my trees file. I changed MrBayes to
output more decimal places on the tree probabilities, because most of my trees were extremely improbable.
I can email you a copy of the unix executable for my modified MrBayes version if you need it (email is
above)


process_trees.data

A file containing one row per tree analysed, in which the interpretation of that tree is given.
There are 4 columns, corresponding to:
	number of protohox genes
	number of parahox gene loss events
	number of Hox gene loss events
	whether it's 'True' that you can reject the TD hypothesis

You will also notice a file called 'current_chunk', which you can delete. It was just a dumb way of 
making the program run faster. 

***************************************************************************
USE

Install Python and BioPython
Run your MrBayes analysis, and use 'sumt' to generate a '.trprobs' file.
Change the INPUT_FILENAME below to match whatever your file is called.
Put a copy of this script into the same directory as your .trprobs file.
Open a terminal, and type

python process_trees.py

then hit enter, and wait.

that's it. 
'''


from Bio.Nexus import Nexus, Nodes
import string

############# STUFF YOU SHOULD LOOK AT BEFORE STARTING ANALYSIS ################

#first define the outgroup, in this case all my trees use evx as an outgroup
OUTGROUP = ['Hro_eve', 'Bfl_EvxB', 'Bfl_EvxA', 'Dme_eve', 'Mmu_Evx1', 'Mmu_Evx2', 'Ttr_eve']

#define the list of ParaHox genes in your dataset
PARAHOX_GENES = ['Mmu_Gsh1', 'Mmu_Gsh2', 'Nvi_Gsx', 'Bfl_Gsh', 'Bfl_Gsx', 'Csp_Gsx', 'Dme_ind', 'Mmu_Xlox', 'Bfl_Xlox', 'Cca_Xlox', 'Csp_Xlox', 'Pst_Xlox', 'Mmu_Cdx1', 'Mmu_Cdx2', 'Csp_Cdx', 'Bfl_Cdx', 'Dme_cad', 'Nvi_Cdx']

#define a dictionary that has all the groupings of hox and parahox genes in it
#NB I call all the central group (6, 7, and 8) '6'
#and all the posterior group (9+), 9
#parahox are called 'G', 'C', and 'X'
GENE_LOOKUP = {'Nvi_Hox1': 1, 'Cva_Hox1': 1, 'Bfl_Hox1': 1, 'Mmu_HoxA1': 1, 'Mmu_HoxB1': 1,
'Mmu_HoxA2': 2, 'Mmu_HoxB2': 2, 'Bfl_Hox2': 2, 'Cva_Hox2': 2, 'Alo_pb': 2,
'Mmu_HoxA3': 3, 'Mmu_HoxB3': 3, 'Bfl_Hox3': 3, 'Cva_Hox3': 3, 'Nvi_Hox3': 3,
'Mmu_HoxA4': 4, 'Mmu_HoxB4': 4, 'Bfl_Hox4': 4, 'Htr_Lox18': 4, 'Nvi_Dfd': 4,
'Mmu_HoxA5': 5, 'Mmu_HoxB5': 5, 'Bfl_Hox5': 5, 'Lan_scr': 5, 'Nvi_scr': 5,
'Mmu_HoxA6': 6, 'Mmu_HoxB6': 6, 'Bfl_Hox6': 6, 'Mmu_HoxB7': 6, 'Mmu_HoxA7': 6, 'Bfl_Hox7': 6, 'Mmu_HoxB8': 6, 'Mmu_HoxC8': 6, 'Bfl_Hox8': 6, 'Nvi_Lox5': 6, 'Nvi_Lox2': 6, 'Nvi_Lox4': 6, 'Hme_Lox4': 6, 'Hro_Lox2': 6, 'Hro_Lox5': 6,
'Mmu_HoxA9': 9, 'Mmu_HoxB9': 9, 'Bfl_Hox9': 9, 'Lan_Post1': 9, 'Nvi_Post1': 9, 'Mmu_HoxA10': 9, 'Mmu_HoxC10': 9, 'Bfl_Hox10': 9, 'Lan_Post2': 9, 'Nvi_Post2': 9,
'Mmu_Gsh1': 'G', 'Mmu_Gsh2': 'G', 'Bfl_Gsx': 'G', 'Csp_Gsx': 'G',
'Mmu_Cdx1': 'C', 'Mmu_Cdx2': 'C', 'Bfl_Cdx': 'C', 'Csp_Cdx': 'C', 'Nvi_Cdx': 'C',
'Mmu_Xlox': 'X', 'Bfl_Xlox': 'X', 'Csp_Xlox': 'X', 'Pst_Xlox': 'X'}

#two possible gene orders, allowing for the fact that the parahox cluster might have been inverted at some point
#when it split from the hox cluster, if the whole thing was formed by tandem duplication
GENE_ORDER1 = [1, 2, 3, 4, 5, 6, 9, 'C', 'X', 'G']
GENE_ORDER2 = [1, 2, 3, 4, 5, 6, 9, 'G', 'X', 'C']

GENE_ORDER_LENGTH = 10

#filenames
INPUT_FILENAME = 'Dataset2_uncon.nex.trprobs'
OUTPUTFILENAME = 'process_trees.out'
DATA_FILENAME = 'process_trees.data'

#how many trees to analyse? IF you want to analyse ALL the trees in your input file
#set this to zero. Otherwise, if you want ot analyse the 95% confidence set, 
#put in the number of trees in that set
TREES_TO_ANALYSE = 0

#if you are using a certain confidence set of trees, set this equal to that set
#this allows the script to correct the posterior odds as it goes
#(e.g. if using the 95% confidence set, you need to multiply all posterior
#odds by 100/95.
PERCENTAGE = 100


################################################################################

#set up the results list, this will be a list of Result objects (see class Result())
HOX_RESULTS = []
PROB_RESULTS = []
RESULTS = []

#these are the lists i use to chop up big file (>2000 trees) into manageable chunks
HEADER = []
TREES = []
END = 'end;'

#this is a dictionary which will eventually contain all the hypotheses (of protohox etc)
#and their cumulative probabilities
HYPOTHESES = {}

#print a star every 'star' trees processed
star = 1000

#an annoying variable i had to make global
NODE_WITHOUT_OUTGROUPS = 123

#define a results class where I can store various bits of stuff and grab them out later
class Result():
	def __init__(self, tree_name, p_value, protohox, parahox_loss, hox_loss, violates_tandem):
		self.tree_name = tree_name					#name of the tree from the mrbayes file
		self.p_value = p_value						#posterior prob. of tree from mybayes file
		self.protohox = protohox					#number of protohox genes tree supports
		self.parahox_loss = parahox_loss			#number of parahox losses tree supports
		self.hox_loss = hox_loss					#number of hox losses tree supports
		self.violates_tandem = violates_tandem		#violates tandem duplication? (True/False)

#these are some sets that get used later on, to tell me if a node contains
#all hox, all parahox, or a mixture of the two
#what are the sets I'm interested in?
All_Hox = set([False])
All_ParaHox = set([True])
Mixture = set([True,False])


###########################Here are the functions###############################

#this nifty yield function turns a recursion into an iteration.
#it moves down the tree, and gives you one node at a time, without putting
#any annoying shit on the stack. Neat-o. it takes a NODE OBJECT and a TREE OBJECT.
def daughter_nodes(node, current_tree):
	
	yield node
	
	for daughter_id in node.get_succ():
		
		daughter = current_tree.node(daughter_id)
		
		for daughter2 in daughter_nodes(daughter, current_tree):
			
			yield daughter2
			
#this function opens up your input file, and looks for the posterior probability
#of each tree that's there.
def extract_posterior_probs():
	
	print '\n getting posterior probabilities from input file'
	
	#open up the file
	inputfile = open(INPUT_FILENAME, "r")
	
		
	#get out each line in the file
	for line in inputfile.readlines():
	
				
		#split it into words separated by whitespace
		words = string.split(line)
				
		#if there are any words at all, and the first word is 'tree' then it's a tree line
		if words and words[0] == 'tree':
		
			this_tree_name = words[1]
			
			#the fifth word should be posterior probability, but the last character
			#is always a comma, so lop that off
			this_p_value = words[4][:-1]
			
			#now multiply it by a factor which corrects it for the confidence set of trees you are using
			#e.g. if using the 95% confidence set, you need to multiply all odds by 100/95
			this_p_value = float(this_p_value) * (100.0/float(PERCENTAGE))
		
			PROB_RESULTS.append([this_tree_name, this_p_value])
	
	inputfile.close()
	
	print "\nfinished extracting posterior probabilities. \n"

# pass this function the root node and a results object
#it takes a NODE OBJECT and a TREE OBJECT
def process_nodes(node, current_tree, current_tree_result): 

	#first, see if this tree violates the tandem duplication scenario
	violates_tandem_duplication = False #start by assuming it doesn't
	#now run through all nodes

	#this is a dict with entries for each orthology group
	#it is used to figure out whether orthology groups are monophyletic or not
	#we start off assuming they're not
	#and mark them as monophyletic once we see that they are
	monophyletic = {1: False, 2: False, 3: False, 4: False, 5: False, 6: False, 9: False, 'C': False, 'X': False, 'G': False}
	
	
	
	for daughter in daughter_nodes(node, current_tree):
		(violates_tandem_duplication, monophyletic) = test_node_for_tandem_duplication(daughter, current_tree, monophyletic)
		
		#if you find out that the tree does violate the tandem duplication, then stop looking
		#since only one node needs to break the conditions for it to be false.
		if violates_tandem_duplication == True:
			
			break
						
	#now put the answer for this tree into the results
	current_tree_result[3] = violates_tandem_duplication	
			
	current_tree_result = get_protohox_hypothesis(node, current_tree, current_tree_result)
	
	#print 'the results for this tree are:', current_tree_result
	
	return current_tree_result



# pass this function the root node object, a tree object and a results list
def get_protohox_hypothesis(node, current_tree, current_tree_result): 

	
	# test_node looks at the daughters of a node to see if they have just hox 
	# just parahox, or a mixture. If it's a mixture then they
	#themselves need testing, they'll be put in daughters_to_test
	daughters_to_test = test_node(node, current_tree, current_tree_result)

	#now recursively call the function to test the successors
	#this will run through the whole tree putting each of the daughters_to_test 
	#through the test_node function	
	for daughter in daughters_to_test:
	
		#get the node object of the daughter in question and pass it for processing
		daughter_node = current_tree.node(daughter)
		get_protohox_hypothesis(daughter_node, current_tree, current_tree_result)
		
	return current_tree_result


#moves up the tree until we find the node without outgroups
def remove_outgroups(node, current_tree):
	
	for daughter in daughter_nodes(node, current_tree):
		
		daughter_id = daughter.get_id()
		
		tips = current_tree.get_taxa(daughter_id)
		
		#if the number of tips is ever less than the number of genes in GENE_LOOKUP then something's wrong, 
		#most probably this would occur if the outgroup wasn't monophyletic
		if len(tips) < len(GENE_LOOKUP):
			print '\nERROR: looks like your outgroup is not monophyletic\n'
		
		contains_outgroup = set([gene in OUTGROUP for gene in tips])
		
		#find the FIRST node without any outgroups in it
		if contains_outgroup == set([False]):
			node_i_want = daughter
			break
	else: #if you didn't find it, something's very wrong
		print '\nERROR: failed to remove outgroups - could not find the node with just Hox and ParaHox, and no outgroups\n'
	
	return node_i_want
	
	
#takes as input a node, and the results structure then looks at the two
#daughters of that node to see if they are all Hox or all Parahox 
def test_node(node, current_tree, current_tree_result):
	
	#define the daughters_to_test() list, which will be retured at the end
	#with any of the daughters which need further testing
	daughters_to_test = []
	
	#this list will hold the identity of each daughter node - whether it's 
	#hox, parahox or mixture. This allows me to work out the results at the end
	daughters_test_results = []

	#get daughters of this node as integers
	daughters = node.get_succ()

	#for each daughter node
	for daughter in daughters:

		#get all tips deriving from the node of interest
		tips = current_tree.get_taxa(daughter)
		
		#if you get this far, then there are only hox and parahox genes deriving
		#from the node we've got. Question is, is it hox, parahox, or both
		node_classification = set([gene in PARAHOX_GENES for gene in tips])

		#if it's a mixture, send it back for processing and don't do anything
		if node_classification == Mixture:
			
			daughters_to_test.append(daughter)
			daughters_test_results.append(1)
								
		#if it's just hox, then count that as one protohox gene (result[0])
		#and one parahox gene loss (result[2])
		if node_classification == All_Hox:		
			daughters_test_results.append(2)

			
		#if it's just parahox, then count that as one protohox gene (result[0])
		#and one hox gene loss (result[2])
		if node_classification == All_ParaHox:						
			daughters_test_results.append(3)

	#sort out the results.
	#if the test results contain '1' AND '2' it's daugters are 'mixture' and 'hox' so add one protohox gene, and one parahox loss
	#if the test results contain '1' AND '3' it's daugters are 'mixture' and 'parahox' so add one protohox gene, and one hox loss
	#if the test results contain '3' AND '2' it's daugters are 'parahox' and 'hox' so add one protohox gene

	if 1 in daughters_test_results and 2 in daughters_test_results:	
		current_tree_result[0] = current_tree_result[0] + 1
		current_tree_result[1] = current_tree_result[1] + 1

	if 1 in daughters_test_results and 3 in daughters_test_results:	
		current_tree_result[0] = current_tree_result[0] + 1
		current_tree_result[2] = current_tree_result[2] + 1

	if 2 in daughters_test_results and 3 in daughters_test_results:	
		current_tree_result[0] = current_tree_result[0] + 1

		
	return daughters_to_test

def test_node_for_tandem_duplication(node, current_tree, monophyletic):

	"""In order to fit the tandem duplication hypothesis, a couple of things must be true about the tree, these are:
	1. All descendents of a given node must form a contiguous set (i.e. a slice) from GENEORDER1 or GENEORDER2
	2. The first time you see any particular orthology group on its own, it must have all members of that group in it
	   i.e. all orthology groups have to be monophyletic.
	"""

	node_id = node.get_id()
	tips = current_tree.get_taxa(node_id)
	
	#classify the set of tips into which gene groups they belong
	gene_group_list = []
	for tip in tips:
		gene_group_list.append(GENE_LOOKUP[tip])
	
	#now make a set out of that list
	#the length of this set tells you how many different orthology groups are represented in the descendents of this node
	gene_group_set = set([])
	for group in gene_group_list:
		gene_group_set.add(group)
	
	#the length of the slice you're interested in will just be the length of the groups set
	slice_length = len(gene_group_set)
	
	#if the set is bigger than 1 (i.e. if there is more than one orthologue represented
	#in all the tips), then you have to test whether those genes make a contiguous set within the gene order
	#to do this, just slice up GENE_ORDER1 and 2, and see if the gene_group_set corresponds to one of those slices
	if slice_length > 1:
		
		#what's the maximum number of slices of slice_length you can get from the gene_order lists
		max_slices = GENE_ORDER_LENGTH + 1 - slice_length
		
		#now run through each possible slice
		for i in range(max_slices):
			
			#make a set for each gene order option
			this_slice1 = set([])
			this_slice2 = set([])
		
		
			#and build your slices
			for item in range(slice_length):
			
				this_slice1.add(GENE_ORDER1[i+item])
				this_slice2.add(GENE_ORDER2[i+item])

			#now test whether your actual set matches either of these
			#if it does, the length of the 'differences' set will be zero at least once
			test1 = gene_group_set.difference(this_slice1)
			test2 = gene_group_set.difference(this_slice2)
			
			if len(test1) == 0 or len(test2) == 0:
				
				#if either of those is true, then your node conforms to the tandem duplication scenario,
				#so you don't need to test anymore slices, i.e. it only need conform to one slice
				violates_tandem_duplication = False
				break
					
			else: #otherwise, it doesn't so disregard the rest of the tree for tandem duplications
				violates_tandem_duplication = True
				
	
	else: #if the slice length is only 1, then it's only got one orthology group in
		
		
		#see if this orthology group is marked as monophyletic already
		#if it is, then don't do anything else
		this_group = gene_group_set.pop()
		
		
		is_monophyletic = monophyletic.get(this_group)
		
		
		#if it's not already labelled as monophyletic, then test if it is
		if is_monophyletic == False:
		
			#are all the members of the set there
			#get the values out of the gene-lookup table
			values = GENE_LOOKUP.values()
			
			#and see if the number of times the group in question appears in 
			#gene_lookup (i.e. the number of orthologues of this_group) is equal to the number of tips from this node
			#since we see deeper nodes first, if a particular orthology group is monophyletic, then this 
			#condition will be satisfied the first time we see that group
			#and since we never reverse a decision on monophyly, we can keep looking at nodes within this group later
			#without worrying about this particular condition.
			
			if values.count(this_group) == len(tips):
				
				#set the dict value to monophyletic
				monophyletic[this_group] = True
				violates_tandem_duplication = False
		
		
			else: #if it's not monophyletic after you test it, then the tree doesn't fit the tandem duplication hypothesis 
				
				violates_tandem_duplication = True
				
		
		else: #it must be monophyletic, so it's all OK for this node
			violates_tandem_duplication = False
	
	return (violates_tandem_duplication, monophyletic)


def process_tree():

	print "starting to process trees...\n"

	#open your INPUT file
	print "opening nexus file using biopython Nexus module"
	tree_file = Nexus.Nexus('current_chunk')
	print "finished opening file\n"

	print 'begin figuring out trees...'
	
	datafile = open(DATA_FILENAME, "a")
	
	#for each tree in the tree file
	for current_tree in tree_file.trees:
		
				
		#Arbitrarily root the tree with one of the outgroup
		root_node = current_tree.root_with_outgroup(OUTGROUP[0])
	
		#now get the node object
		node = current_tree.node(root_node)
				
		#now move up the tree until you find the node with just hox and parahox, and no outgroups.
		#this is returned as a NODE OBJECT
		node_without_outgroup = remove_outgroups(node, current_tree)
		
		#now pass that node object to the process_node function, which will treat
		#it as the root node, and spit out how many hox, parahox and protohox you have
		
		current_tree_result = [0,0,0,False]
		process_nodes(node_without_outgroup, current_tree, current_tree_result)
					
		datafile.write('%d\t%d\t%d\t%s\n' % (current_tree_result[0], current_tree_result[1], current_tree_result[2], current_tree_result[3]))
	
	print 'done figuring out trees...'
	
	datafile.close()

#this function will take an input file and chop it up into manageable chunks for the 
#rest of the program
def make_chunk(chunk_size, chunk_number):


	print "making a chunk"

	#this is just a list of strings which will constitute the trees of the file
	trees_in_chunk = []

	start_tree = chunk_size * chunk_number
	
	print 'start_tree =', start_tree
		
	trees_in_chunk = TREES[start_tree:(start_tree+chunk_size)]

	print 'number of trees in this chunk =', len(trees_in_chunk)
	
	#open up a file
	chunk_file = open('current_chunk', 'w')
	
	#put all the stuff in it
	for line in HEADER:
		chunk_file.write('%s' % line)
	for tree in trees_in_chunk:
		chunk_file.write('%s' % tree)
	chunk_file.write('%s' %END)
	
	chunk_file.close()
	

def build_arrays_for_chunking():

	#open up the file
	inputfile = open(INPUT_FILENAME, "r")

	i = 1
	
	#get out each line in the file
	for line in inputfile.readlines():
	
		
		#split it into words separated by whitespace
		words = string.split(line)
		
		#build the header from everything other than the trees and the end
		if words and words[0] != 'tree' and words[0] != 'end;':
			
			HEADER.append(line)
		
		#if there are any words at all, and the first word is 'tree' then it's a tree line
		if words and words[0] == 'tree':
		
			TREES.append(line)
			
			if TREES_TO_ANALYSE != 0 and i == TREES_TO_ANALYSE:
				break
			i = i + 1

	inputfile.close()

	return len(TREES)

###############################################################################

################################ Results calculations ##################
def do_clever_stuff():

	print '\n**************** PROCESSING RESULTS ******************\n'

	#get all the posterior probabilities from the input file
	extract_posterior_probs()

	#now open the file with all the individual tree results in
	datafile = open(DATA_FILENAME, "r")
	
	i = 0
	
	#get out each line in the file
	for line in datafile.readlines():
				
		#split it into words separated by whitespace
		words = string.split(line)
					
		current_result = Result(PROB_RESULTS[i][0], PROB_RESULTS[i][1], int(words[0]), int(words[1]), int(words[2]), words[3])
		
		#print RESULTS[i].tree_name, RESULTS[i].p_value, RESULTS[i].protohox, RESULTS[i].parahox_loss, RESULTS[i].hox_loss, RESULTS[i].violates_tandem
	
		#now this result is in a nice Result object
	
		###########   Buliding Cumulative p_values for each hypothesis ###########
		#define a 'hypothesis' as the number of protohox genes suggested by the tree analysis
		#then look for it in a dictionary of hypotheses
		#if that hypothesis is already there, add the current p_value to it
		#current_hypothesis has to be a string, as lists are unhashable
		current_hypothesis = '%d\t%d\t%d' % (current_result.protohox, current_result.parahox_loss, current_result.hox_loss)
		
		c = HYPOTHESES.setdefault(current_hypothesis, 0.0)
	
		p = float(current_result.p_value)
	
		HYPOTHESES[current_hypothesis] = c + p
	
		############# cumulative probability for tandem duplications ##############
		
		tandem_hypothesis = '%s' % current_result.violates_tandem
		
		c = HYPOTHESES.setdefault(tandem_hypothesis, 0.0)
		
		p = float(current_result.p_value)
	
		HYPOTHESES[tandem_hypothesis] = c + p

		i = i + 1
	
	datafile.close()

	#now you have a dict called HYPOTHESES which has all the hyoptheses and their cumulative probabilities
	#open the output file, and dump this information into it
	output_file = open(OUTPUTFILENAME, 'w')
	
	output_file.write('%d trees were processed in total\n' % len(PROB_RESULTS))
	
	output_file.write('and %d different hypotheses are represented\n\n' % len(HYPOTHESES))
	
	output_file.write('protohox\tparahox loss\thox loss\tcumulative p\n')
	
	for hypothesis in HYPOTHESES:
					
		output_file.write('%s\t%f\n' % (hypothesis, HYPOTHESES[hypothesis]))
	
	output_file.close()
	
	
	print '\n _______________________DONE________________________________ \n'






def run_the_program():

	chunk_size = 1000
	
	print "\nStarting analysis\n"
	
	#first build the arrays you need to make smaller files
	number_of_trees = build_arrays_for_chunking()
	
	print '#trees =', number_of_trees
	
	#the number of small files (chunks) will be the number of trees / chunk size, rounded up
	number_of_chunks = (number_of_trees / chunk_size)
	if number_of_trees % chunk_size > 0:
		number_of_chunks = number_of_chunks + 1
	
	print '#chunks =', number_of_chunks
	
	#this just opens the DATA FILE and then closes it, to wipe anything that was in it already.
	#this saves dumb mistakes of appending one set of results on top of another
	datafile = open(DATA_FILENAME, "w")
	datafile.close()

	
	for chunk_number in range(number_of_chunks): 
		
		print '\n################## chunk_number %d of %d ##################\n' % ((chunk_number+1), number_of_chunks)
		
		make_chunk(chunk_size, chunk_number)
		
		process_tree() #now get all the hypotheses for each tree in that chunk
	
	
	do_clever_stuff()
	


run_the_program()





